import json
import pandas as pd
from pathlib import Path
from itertools import repeat
from collections import OrderedDict


def ensure_dir(dirname):
    dirname = Path(dirname)
    if not dirname.is_dir():
        dirname.mkdir(parents=True, exist_ok=False)

def read_json(fname):
    fname = Path(fname)
    with fname.open('rt') as handle:
        return json.load(handle, object_hook=OrderedDict)

def write_json(content, fname):
    fname = Path(fname)
    with fname.open('wt') as handle:
        json.dump(content, handle, indent=4, sort_keys=False)

def inf_loop(data_loader):
    ''' wrapper function for endless data loader. '''
    for loader in repeat(data_loader):
        yield from loader

class MetricTracker:
    def __init__(self, *keys, writer=None):
        self.writer = writer
        self._data = pd.DataFrame(index=keys, columns=['total', 'counts', 'average'])
        self.reset()
        
    def reset(self):
        for col in self._data.columns:
            self._data[col].values[:] = 0

    def update(self, key, value, n=1):
        if self.writer is not None:
            self.writer.add_scalar(key, value)
        self._data.total[key] += value * n
        self._data.counts[key] += n
        self._data.average[key] = self._data.total[key] / self._data.counts[key]

    def avg(self, key):
        return self._data.average[key]
    
    def result(self):
        return dict(self._data.average)



import numpy as np
import torch.nn as nn
from abc import abstractmethod


class BaseModel(nn.Module):
    """
    Base class for all models
    """
    @abstractmethod
    def forward(self, *inputs):
        """
        Forward pass logic

        :return: Model output
        """
        raise NotImplementedError

    def __str__(self):
        """
        Model prints with number of trainable parameters
        """
        model_parameters = filter(lambda p: p.requires_grad, self.parameters())
        params = sum([np.prod(p.size()) for p in model_parameters])
        return super().__str__() + '\nTrainable parameters: {}'.format(params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
File: convert_onnx.py
Created Date: January 1st 2020
Author: ZL Deng <dawnmsg(at)gmail.com>
---------------------------------------
Last Modified: 6th December 2020 10:58:42 pm
'''

import click
import torch
import warnings
import numpy as np
from os import path
from ribodetector.model import model_cpu as module_arch
from ribodetector.parse_config import ConfigParser


@click.command()
@click.argument('pth', type=click.Path(exists=True))
@click.option('-l', '--length', type=int, default=100, help='The input read length')
def convert_to_onnx(pth, length):
    cd = path.dirname(path.abspath(__file__))
    config_file = path.join(cd, 'config.json')
    config = ConfigParser.from_json(config_file)
    logger = config.get_logger('convert', 1)
    model = config.init_obj('arch', module_arch)
    state = torch.load(pth, map_location=torch.device('cpu'))
    state_dict = state['state_dict']
    model.load_state_dict(state_dict)
    model.eval()
    vocabulary = list('ACGT')
    encodes = np.eye(len(vocabulary))
    x = encodes[np.random.choice(encodes.shape[0], size=length)]
    input_tensor = torch.FloatTensor(np.expand_dims(x, axis=0))

    exported_onnx_file = pth.rsplit(
        '.', 1)[0] + '.onnx'

    logger.info('Converting to ONNX model: {}'.format(exported_onnx_file))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        torch.onnx.export(model,
                          input_tensor,
                          exported_onnx_file,
                          export_params=True,
                          opset_version=10,
                          do_constant_folding=True,
                          input_names=["input"],
                          output_names=["output"],
                          dynamic_axes={"input": {0: "batch_size", 1: "sequence"},
                                        "output": {0: "batch_size", 1: "sequence"}})


if __name__ == '__main__':
    convert_to_onnx()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
File: detect.py
Created Date: January 1st 2020
Author: ZL Deng <dawnmsg(at)gmail.com>
---------------------------------------
Last Modified: 6th December 2020 10:57:59 pm
'''

import os
import gzip
import math
import torch
import argparse

from tqdm import tqdm
from functools import partial
from ribodetector.model import model as module_arch
from collections import defaultdict
from ribodetector.parse_config import ConfigParser
from torch.multiprocessing import Pool
from torch.utils.data import DataLoader
from ribodetector import __version__
from argparse import RawTextHelpFormatter
from ribodetector.data_loader import seq_encoder as SeqEncoder
from ribodetector.data_loader.dataset import SeqData, PairedReadData
from torch.nn.utils.rnn import pad_sequence, pack_sequence

# Get the directory of the program
cd = os.path.dirname(os.path.abspath(__file__))


class Predictor:
    """
    Main class of predictor for rRNA, non-rRNA sequences
    """

    def __init__(self, config, args):
        self.config = config
        self.args = args
        self.logger = config.get_logger('predict', 1, self.args.log)
        self.chunk_size = self.args.chunk_size

    def get_state_dict(self):
        """Select the right model file for classification

        Raises:
            RuntimeError: raise error if input sequence length <40
        """

        self.len = self.args.len

        if self.len < 40:
            #             self.logger.error('{}Sequence length is too short to classify!{}'.format(
            #                 colors.FAIL,
            #                 colors.ENDC))
            #             raise RuntimeError(
            #                 "Sequence length must be set to larger than 40.")

            self.logger.info(
                'The accuracy will drop with reads shorter than 40.')

        # High recall model if ensure non-rRNA
        if self.args.ensure == 'norrna':
            model_file_ext = 'recall'
        else:
            model_file_ext = 'mcc'

        self.state_file = os.path.join(
            cd, self.config['state_file'][model_file_ext])
        # self.logger.info('Using high {} model file: {}{}{}{}'.format(model_file_ext.upper(),
        #                                                             colors.BOLD,
        #                                                             colors.OKCYAN,
        #                                                             self.state_file,
        #                                                             colors.ENDC))

        self.logger.info('Using high {} model'.format(model_file_ext.upper()))
        
        self.logger.info('Log file: {}'.format(
            self.args.log
            ))

    def load_model(self):
        """Load the model onto CUDA device

        Raises:
            RuntimeError: raise error if CUDA device not available
        """
        if self.args.deviceid is not None:
            os.environ["CUDA_VISIBLE_DEVICES"] = self.args.deviceid
        self.get_state_dict()
        model = self.config.init_obj('arch', module_arch)

        if self.config['n_gpu'] > 1:
            model = torch.nn.DataParallel(model)

        if torch.cuda.is_available():
            self.device = 'cuda'
            self.has_cuda = True
            state = torch.load(self.state_file)
        else:
            self.logger.error('{}No visible CUDA devices!{} Please use ribodetector_cpu to run it on CPU if you do not have GPU or \nyou need to install GPU version of PyTorch'.format(
                colors.FAIL,
                colors.ENDC))
            raise RuntimeError(
                "Set CUDA_VISIBLE_DEVICES or use CPU inference.")
        self.logger.info('Model using {} for read length {}{}{}{} loaded'.format(
            self.device,
            colors.BOLD,
            colors.OKCYAN,
            self.len,
            colors.ENDC))

        state_dict = state['state_dict']
        model.load_state_dict(state_dict)

        self.model = model.to(self.device, non_blocking=self.has_cuda)
        self.model.eval()

    def run(self):
        """
        Load data and run the predictor
        """

        num_nonrrna = 0
        num_rrna = 0

        if self.is_paired:
            # Load paired end read files using multiprocessing
            with Pool(2) as p:
                paired_reads = p.map(SeqEncoder.load_reads, self.input)

            paired_reads_data = PairedReadData(paired_reads)
            num_seqs = len(paired_reads_data)

            self.logger.info('{}{}{}{} sequences loaded!'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_seqs,
                colors.ENDC))

            # Open output file handles
            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))
                rrna1_fh = open_for_write(self.rrna[0])
                rrna2_fh = open_for_write(self.rrna[1])
                # num_rrna = 0

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))

            norrna1_fh = open_for_write(self.output[0])
            norrna2_fh = open_for_write(self.output[1])

            if self.args.ensure == 'both':
                # Open the file for writing unclassified reads when ensure is `both`
                unclf1 = self.output[0] + '.unclassified.gz'
                unclf2 = self.output[1] + '.unclassified.gz'
                unclf1_fh = open_for_write(unclf1)
                unclf2_fh = open_for_write(unclf2)
                self.logger.info('Writing unclassified sequences into file: {}{}, {}{}'.format(
                    colors.OKYELLOW,
                    unclf1,
                    unclf2,
                    colors.ENDC))
                num_unknown = 0

            num_batches = math.ceil(num_seqs / self.batch_size)

            data_loader = tqdm(DataLoader(paired_reads_data,
                                          num_workers=self.args.threads,
                                          pin_memory=self.has_cuda,
                                          batch_size=self.batch_size,
                                          collate_fn=self.read_collate_fn),
                               total=num_batches)

            with torch.no_grad():
                for r1, r1_data, r2, r2_data in data_loader:
                    r1_output = self.model(r1_data.to(
                        self.device, non_blocking=self.has_cuda))
                    r2_output = self.model(r2_data.to(
                        self.device, non_blocking=self.has_cuda))

                    r1_dict, r2_dict = self.separate_paired_reads(
                        r1, r1_output, r2, r2_output)
                    
                    num_nonrrna += len(r1_dict[0])
                    num_rrna += len(r1_dict[1])
                    
                    if r1_dict[0]:
                        norrna1_fh.write('\n'.join(r1_dict[0]) + '\n')
                        norrna2_fh.write('\n'.join(r2_dict[0]) + '\n')
                    if self.rrna is not None and r1_dict[1]:
                        rrna1_fh.write('\n'.join(r1_dict[1]) + '\n')
                        rrna2_fh.write('\n'.join(r2_dict[1]) + '\n')

                    if self.args.ensure == 'both' and r1_dict[-1]:
                        unclf1_fh.write('\n'.join(r1_dict[-1]) + '\n')
                        unclf2_fh.write('\n'.join(r2_dict[-1]) + '\n')
                        num_unknown += len(r1_dict[-1])

                    # del r1_data, r2_data, r1_output, r2_output, r1_batch_labels, r2_batch_labels

            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_seqs,
                        colors.ENDC))
            
            # Write predicted rRNA sequences if the rRNA output file is given
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))
            
            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))

            if self.rrna is not None:
                rrna1_fh.close()
                rrna2_fh.close()

            # Write unclassified sequences if the ensure is `both`
            if self.args.ensure == 'both':
                self.logger.info('Discarded {}{}{}{} unclassified sequences'.format(
                    colors.BOLD,
                    colors.OKCYAN,
                    num_unknown,
                    colors.ENDC
                ))
                unclf1_fh.close()
                unclf2_fh.close()

            norrna1_fh.close()
            norrna2_fh.close()

        # Single end reads
        else:
            reads_data = SeqData(SeqEncoder.load_reads(*self.input))

            num_seqs = len(reads_data)

            self.logger.info('{}{}{}{} sequences loaded!'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_seqs,
                colors.ENDC))

            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))

                rrna_fh = open_for_write(self.rrna[0])
                # num_rrna = 0

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))
            norrna_fh = open_for_write(self.output[0])

            num_batches = math.ceil(num_seqs / self.batch_size)
            data_loader = tqdm(DataLoader(reads_data,
                                          num_workers=self.args.threads,
                                          pin_memory=self.has_cuda,
                                          batch_size=self.batch_size,
                                          collate_fn=self.read_collate_fn),
                               total=num_batches)

            with torch.no_grad():
                for reads, data in data_loader:
                    output = self.model(
                        data.to(self.device, non_blocking=self.has_cuda))
                    batch_labels = torch.argmax(output, dim=1).tolist()
                    separated_reads = Predictor.separate_reads(
                        reads, batch_labels)
                
                    num_nonrrna += len(separated_reads[0])
                    num_rrna += len(separated_reads[1])

                    if separated_reads[0]:
                        norrna_fh.write('\n'.join(separated_reads[0]) + '\n')
                    if self.rrna is not None and separated_reads[1]:
                        rrna_fh.write('\n'.join(separated_reads[1]) + '\n')

                    # del data, output, batch_labels
            
            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_seqs,
                        colors.ENDC))
            
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))
            
            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))

            if self.rrna is not None:
                rrna_fh.close()
            norrna_fh.close()

    def run_with_chunks(self):
        """
        Run the model on input sequence files loaded into chunks to reduce the memory comsumption
        """

        num_read = 0
        num_nonrrna = 0
        num_rrna = 0
        
        if self.is_paired:

            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))

                rrna1_fh = open_for_write(self.rrna[0])
                rrna2_fh = open_for_write(self.rrna[1])

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))
            norrna1_fh = open_for_write(self.output[0])
            norrna2_fh = open_for_write(self.output[1])

            if self.args.ensure == 'both':

                unclf1 = self.output[0] + '.unclassified.gz'
                unclf2 = self.output[1] + '.unclassified.gz'
                unclf1_fh = open_for_write(unclf1)
                unclf2_fh = open_for_write(unclf2)
                self.logger.info('Writing unclassified sequences into file: {}{}, {}{}'.format(
                    colors.OKYELLOW,
                    unclf1,
                    unclf2,
                    colors.ENDC))
                num_unknown = 0

            #num_read = 0

            with torch.no_grad():
                # Load paired end read files into chunks
                for chunk in SeqEncoder.get_pairedread_chunks(*self.input,
                                                              chunk_size=self.batch_size * self.chunk_size):
                    paired_reads_data = PairedReadData(chunk)
                    num_read_chunk = len(paired_reads_data)
                    data_loader = DataLoader(paired_reads_data,
                                             num_workers=self.args.threads,
                                             pin_memory=self.has_cuda,
                                             batch_size=self.batch_size,
                                             collate_fn=self.read_collate_fn)
                    num_read += num_read_chunk
                    for r1, r1_data, r2, r2_data in data_loader:
                        r1_output = self.model(r1_data.to(
                            self.device, non_blocking=self.has_cuda))
                        r2_output = self.model(r2_data.to(
                            self.device, non_blocking=self.has_cuda))

                        r1_dict, r2_dict = self.separate_paired_reads(
                            r1, r1_output, r2, r2_output)
                        num_nonrrna += len(r1_dict[0])
                        num_rrna += len(r1_dict[1])
                        if r1_dict[0]:
                            norrna1_fh.write('\n'.join(r1_dict[0]) + '\n')
                            norrna2_fh.write('\n'.join(r2_dict[0]) + '\n')
                        if self.rrna is not None and r1_dict[1]:
                            rrna1_fh.write('\n'.join(r1_dict[1]) + '\n')
                            rrna2_fh.write('\n'.join(r2_dict[1]) + '\n')

                        if self.args.ensure == 'both' and r1_dict[-1]:
                            unclf1_fh.write('\n'.join(r1_dict[-1]) + '\n')
                            unclf2_fh.write('\n'.join(r2_dict[-1]) + '\n')
                            num_unknown += len(r1_dict[-1])

                    self.logger.info('{}{}{} sequences finished!'.format(
                        colors.OKGREEN,
                        num_read,
                        colors.ENDC))

                    # del r1_data, r2_data, r1_output, r2_output, r1_batch_labels, r2_batch_labels
            
            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_read,
                        colors.ENDC))
            
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))

            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))


            if self.rrna is not None:
                rrna1_fh.close()
                rrna2_fh.close()

            if self.args.ensure == 'both':
                self.logger.info('Discarded {}{}{}{} unclassified sequences'.format(
                    colors.BOLD,
                    colors.OKCYAN,
                    num_unknown,
                    colors.ENDC
                ))
                unclf1_fh.close()
                unclf2_fh.close()

            norrna1_fh.close()
            norrna2_fh.close()

        else:
            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE, 
                    ", ".join(self.rrna), 
                    colors.ENDC))
                rrna_fh = open_for_write(self.rrna[0])
                #num_rrna = 0

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))
            norrna_fh = open_for_write(self.output[0])

            #num_read = 0

            with torch.no_grad():
                # Load single end read file into chunks
                for chunk in SeqEncoder.get_seq_chunks(*self.input,
                                                       chunk_size=self.batch_size * self.chunk_size):
                    reads_data = SeqData(chunk)
                    num_read_chunk = len(reads_data)

                    data_loader = DataLoader(reads_data,
                                             num_workers=self.args.threads,
                                             pin_memory=self.has_cuda,
                                             batch_size=self.batch_size,
                                             collate_fn=self.read_collate_fn)
                    num_read += num_read_chunk

                    for reads, data in data_loader:
                        output = self.model(
                            data.to(self.device, non_blocking=self.has_cuda))
                        batch_labels = torch.argmax(output, dim=1).tolist()
                        separated_reads = Predictor.separate_reads(
                            reads, batch_labels)
                        
                        num_nonrrna += len(separated_reads[0])
                        num_rrna += len(separated_reads[1])
                        
                        if separated_reads[0]:
                            norrna_fh.write(
                                '\n'.join(separated_reads[0]) + '\n')
                        if self.rrna is not None and separated_reads[1]:
                            rrna_fh.write('\n'.join(separated_reads[1]) + '\n')

                    # del data, output, batch_labels

                    self.logger.info('{}{}{} sequences finished!'.format(
                        colors.OKGREEN,
                        num_read,
                        colors.ENDC))

            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_read,
                        colors.ENDC))
            
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))

            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))
            
            if self.rrna is not None:
                rrna_fh.close()
            norrna_fh.close()

    def detect(self):
        """Wrapper for all steps

        Raises:
            RuntimeError: raise error if the number of input read files and output files are invalid
        """
        self.input = self.args.input
        self.output = self.args.output
        self.rrna = self.args.rrna
        self.pack_seq = self.config['arch']['args']['pack_seq']

        num_inputs = len(self.input)
        num_rrna_outputs = None if self.rrna is None else len(self.rrna)
        num_norrna_outputs = len(self.output)
        if num_inputs != num_norrna_outputs or num_inputs > 2:
            self.logger.error('{}The number of input and output sequence files is invalid!{}'.format(
                colors.FAIL,
                colors.ENDC))
            raise RuntimeError(
                "Input or output should have no more than two files and they should have the same number of files.")

        if num_rrna_outputs is not None and num_rrna_outputs != num_inputs:
            self.logger.error('{}The number of output rRNA sequence files is invalid!{}'.format(
                colors.FAIL,
                colors.ENDC))
            raise RuntimeError(
                "Ouput rRNA should have no more than two files and they should the same number with input files.")

        self.is_paired = (num_inputs == 2)

        if self.is_paired:
            # self.batch_size = math.floor(
            #     200 * math.floor(0.9 * self.args.memory) / (4 * self.len)) * 2 * 1024
            batch_size_ = ((self.args.memory - 2) * 1024 *
                           1024) / (2 * self.len * 6.4)
            self.batch_size = 2 ** math.floor(math.log2(batch_size_))
            self.read_collate_fn = partial(
                unlabeled_paired_read_collate_fn, max_len=self.len, pack_seq=self.pack_seq)
        else:
            # self.batch_size = math.floor(
            #     200 * math.floor(0.9 * self.args.memory) / (self.len * 2)) * 2 * 1024
            batch_size_ = ((self.args.memory - 2) *
                           1024 * 1024) / (self.len * 6.4)
            self.batch_size = 2 ** math.floor(math.log2(batch_size_))
            self.read_collate_fn = partial(
                unlabeled_read_collate_fn, max_len=self.len, pack_seq=self.pack_seq)

        self.logger.info('Choose batch size: {}{}{}{} based on the given GPU RAM size {}GB and max read length {}'.format(
            colors.BOLD,
            colors.OKCYAN,
            self.batch_size,
            colors.ENDC,
            self.args.memory,
            self.len
        ))

        if self.chunk_size is None:
            self.run()
        else:
            self.run_with_chunks()

#################################################################################################################
    @staticmethod
    def generate_chunks(reads, n):
        """Yield successive n-sized chunks from read list."""
        for i in range(0, len(reads), n):
            yield reads[i:i + n]

    @staticmethod
    def generate_paired_read_chunks(reads, n):
        """Yield successive n-sized chunks from paired end read list."""
        r1, r2 = reads
        for i in range(0, len(r1), n):
            yield r1[i:i + n], r2[i:i + n]

    @staticmethod
    def separate_reads(reads, labels):
        """Split the input reads based on the predicted label

        Args:
            reads (list): batch of input reads 
            labels (list): list of predicted labels for the input reads

        Returns:
            dict: dict with key being label and value being read
        """
        reads_dict = defaultdict(list)
        for read, label in zip(reads, labels):
            reads_dict[label].append(read)
        return reads_dict

    def separate_paired_reads(self, r1_reads, r1_outs, r2_reads, r2_outs):
        r1_dict = defaultdict(list)
        r2_dict = defaultdict(list)

        if self.args.ensure == 'rrna':
            r1_labels = torch.argmax(r1_outs, axis=1).tolist()
            r2_labels = torch.argmax(r2_outs, axis=1).tolist()
            for r1, r1_label, r2, r2_label in zip(r1_reads, r1_labels, r2_reads, r2_labels):

                if r1_label == r2_label == 1:
                    final_label = 1
                else:
                    final_label = 0
                r1_dict[final_label].append(r1)
                r2_dict[final_label].append(r2)
        elif self.args.ensure == 'norrna':
            r1_labels = torch.argmax(r1_outs, axis=1).tolist()
            r2_labels = torch.argmax(r2_outs, axis=1).tolist()
            for r1, r1_label, r2, r2_label in zip(r1_reads, r1_labels, r2_reads, r2_labels):
                if r1_label == r2_label == 0:
                    final_label = 0
                else:
                    final_label = 1

                r1_dict[final_label].append(r1)
                r2_dict[final_label].append(r2)
        elif self.args.ensure == 'both':
            r1_labels = torch.argmax(r1_outs, axis=1).tolist()
            r2_labels = torch.argmax(r2_outs, axis=1).tolist()
            for r1, r1_label, r2, r2_label in zip(r1_reads, r1_labels, r2_reads, r2_labels):
                if r1_label == r2_label == 0:
                    final_label = 0
                elif r1_label == r2_label == 1:
                    final_label = 1
                else:
                    final_label = -1

                r1_dict[final_label].append(r1)
                r2_dict[final_label].append(r2)
        else:

            final_labels = torch.argmax(r1_outs + r2_outs, axis=1).tolist()
            for r1, r2, final_label in zip(r1_reads, r2_reads, final_labels):

                r1_dict[final_label].append(r1)
                r2_dict[final_label].append(r2)

        return r1_dict, r2_dict


def unlabeled_read_collate_fn(batch, max_len=100, pack_seq=True):
    """Encode a bacth of single end short reads data with given max len

    Args:
        batch (list(tuple)): a batch of single end short reads
        max_len (int, optional): maximum length of the encoded data. Defaults to 100.
        pack_seq (bool, optional): if the output sequence tensor should be packedsequence. Defaults to True.

    Returns:
        tuple(list): read list, encoded read list
    """
    read_list = []
    encoded_read_list = []
    for read in batch:
        read_list.append('\n'.join(read))
        encoded_read_list.append(torch.FloatTensor(
            SeqEncoder.encode_read(read[1][:max_len])))

    if pack_seq:
        read_data = pack_sequence(encoded_read_list, enforce_sorted=False)
    else:
        read_data = pad_sequence(encoded_read_list, batch_first=True)

    return read_list, read_data


def unlabeled_paired_read_collate_fn(batch, max_len=100, pack_seq=True):
    """Encode a bacth of paired end short reads data with given max len

    Args:
        batch (list(tuple)): a batch of paired end short reads
        max_len (int, optional): maximum length of the encoded data. Defaults to 100.
        pack_seq (bool, optional): if the output sequence tensor should be packedsequence. Defaults to True.

    Returns:
        tuple(list): R1 read list, encoded R1 list, R2 read list, encoded R2 list
    """
    r1_list = []
    r2_list = []
    encoded_r1_list = []
    encoded_r2_list = []
    for r1, r2 in batch:
        r1_list.append('\n'.join(r1))
        r2_list.append('\n'.join(r2))
        r1_read = r1[1]
        r2_read = r2[1]

        encoded_r1_list.append(torch.FloatTensor(
            SeqEncoder.encode_read(r1_read[:max_len])))

        encoded_r2_list.append(torch.FloatTensor(
            SeqEncoder.encode_read(r2_read[:max_len])))

    if pack_seq:
        r1_data = pack_sequence(encoded_r1_list, enforce_sorted=False)
        r2_data = pack_sequence(encoded_r2_list, enforce_sorted=False)

    else:
        r1_data = pad_sequence(encoded_r1_list, batch_first=True)
        r2_data = pad_sequence(encoded_r2_list, batch_first=True)
    return r1_list, r1_data, r2_list, r2_data


def open_for_write(read_file):
    """Open a plain text or gzipped text file for writing according to the file name extension

    Args:
        read_file (str): the file name to write

    Returns:
        file handle: the opened file handle for writing
    """
    if read_file.endswith('gz'):
        return gzip.open(read_file, mode='wt', compresslevel=5)
    else:
        return open(read_file, 'w')


class colors:
    """
    Define the color of logger text
    """

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    OKYELLOW = '\033[33m'
    OKMAG = '\033[35m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    UPDATE = '\033[F'


def main():
    args = argparse.ArgumentParser(
        description='rRNA sequence detector', formatter_class=RawTextHelpFormatter)
    args.add_argument('-c', '--config', default=None, type=str,
                      help='Path of config file')
    args.add_argument('-d', '--deviceid', default=None, type=str,
                      help='Indices of GPUs to enable. Quotated comma-separated device ID numbers. (default: all)')
    args.add_argument('-l', '--len', type=int, required=True,
                      help='Sequencing read length. Note: the accuracy reduces for reads shorter than 40.')
    args.add_argument('-i', '--input', default=None, type=str, nargs='*', required=True,
                      help='Path of input sequence files (fasta and fastq), the second file will be considered as second end if two files given.')
    args.add_argument('-o', '--output', default=None, type=str, nargs='*', required=True,
                      help='Path of the output sequence files after rRNAs removal (same number of files as input). \n(Note: 2 times slower to write gz files)')
    args.add_argument('-r', '--rrna', default=None, type=str, nargs='*',
                      help='Path of the output sequence file of detected rRNAs (same number of files as input)')
    args.add_argument('-e', '--ensure', default="none", type=str, choices=['rrna', 'norrna', 'both', 'none'],
                      help='''Ensure which classificaion has high confidence for paired end reads.
norrna: output only high confident non-rRNAs, the rest are clasified as rRNAs;
rrna: vice versa, only high confident rRNAs are classified as rRNA and the rest output as non-rRNAs;
both: both non-rRNA and rRNA prediction with high confidence;
none: give label based on the mean probability of read pair.
      (Only applicable for paired end reads, discard the read pair when their predicitons are discordant)''')

    args.add_argument('-t', '--threads', default=10, type=int,
                      help='Number of threads to use. (default: 10)')
    args.add_argument('-m', '--memory', default=32, type=int,
                      help='Amount (GB) of GPU RAM. (default: 12)')
    args.add_argument('--chunk_size', default=None, type=int,
                      help='Use this parameter when having low memory. Parsing the file in chunks.\n{}.\n{}.'.format(
                          'Not needed when free RAM >=5 * your_file_size (uncompressed, sum of paired ends)',
                          'When chunk_size=256, memory=16 it will load 256 * 16 * 1024 reads each chunk (use ~20 GB for 100bp paired end)'
                      ))
    args.add_argument('--log', default=None, type=str,
                      help='Log file name')
    args.add_argument('-v', '--version', action='version',
                      version='%(prog)s {version}'.format(version=__version__))

    if not isinstance(args, tuple):
        args = args.parse_args()
    if args.config is None:
        config_file = os.path.join(cd, 'config.json')
    else:
        config_file = args.config
    config = ConfigParser.from_json(config_file)
    seq_pred = Predictor(config, args)
    seq_pred.load_model()
    seq_pred.detect()


if __name__ == '__main__':
    main()



import torch
from torch import jit
import torch.nn as nn
from torch import Tensor
from ribodetector.base import BaseModel


class SeqModel(BaseModel):
    def __init__(self, input_size,
                 hidden_size,
                 num_layers,
                 num_classes,
                 batch_first=True,
                 bidirectional=True,
                 pack_seq=False):
        super(SeqModel, self).__init__()
        self.rnn = nn.LSTM(
            input_size=input_size,      # width of input
            hidden_size=hidden_size,     # number of rnn hidden unit
            num_layers=num_layers,       # number of RNN layers
            batch_first=batch_first,
            bidirectional=bidirectional,
        )
        self.hidden_size = hidden_size
        self.out = nn.Linear(hidden_size * 2, num_classes)    # output layer

        self.forward = self.forward_last

    def forward_last(self, x):
        r_out, _ = self.rnn(x, None)

        # last_out = r_out[torch.arange(
        #     x.size(0)), (x.sum(2) != 0).cumsum(1).argmax(1), :]

        last_out = last_out_items(x, r_out)
        out = self.out(last_out)
        return out

    # If there is no ambigous nucleotide in the sequence
    def forward_fast(self, x):
        r_out, _ = self.rnn(x, None)

        last_out = r_out[torch.arange(
            x.size(0)), x.sum((1, 2)).long() - 1, :]
        out = self.out(last_out)
        return out

    def forward_lastforward_firstreverse(self, x):
        r_out, (h_n, _c_n) = self.rnn(x, None)

        last_out = lastforward_firstreverse_out_items(
            x, r_out, h_n, self.hidden_size)
        out = self.out(last_out)
        return out


@jit.script
def last_out_items(ts: Tensor, r_out: Tensor) -> Tensor:
    last_out = r_out[torch.arange(ts.size(0)), ts.size(
        1) - 1 - ts.sum(2).flip(1).argmax(1), :]

    return last_out


# The last forward unpad and first reverse rnn out or reverse hn
@jit.script
def lastforward_firstreverse_out_items(ts: Tensor,
                                       r_out: Tensor,
                                       h_n: Tensor,
                                       hidden_size: int) -> Tensor:
    last_out = torch.cat((r_out[torch.arange(ts.size(0)), ts.size(
        1) - 1 - ts.sum(2).flip(1).argmax(1), :hidden_size], h_n[1]), 1)

    return last_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
File: detect_cpu.py
Created Date: January 1st 2020
Author: ZL Deng <dawnmsg(at)gmail.com>
---------------------------------------
Last Modified: 7th March 2022 12:02:19 pm
'''

import os
import math
import gzip
import argparse
import platform
import itertools
import onnxruntime
import numpy as np
from tqdm import tqdm
import multiprocessing as mp
import multiprocessing.util
from collections import defaultdict
from ribodetector import __version__

from argparse import RawTextHelpFormatter
from ribodetector.parse_config import ConfigParser
import ribodetector.data_loader.seq_encoder as SeqEncoder

# Get the directory of the program
cd = os.path.dirname(os.path.abspath(__file__))

## makes the socket addresses extremely random again to esure no address conflicts
multiprocessing.util.abstract_sockets_supported = False

class Predictor:
    """
    Main class of predictor for rRNA, non-rRNA sequences
    """

    def __init__(self, config, args):
        self.config = config
        self.args = args
        self.logger = config.get_logger('predict', 1, self.args.log)
        self.chunk_size = self.args.chunk_size
        # self.input = self.args.input
        # self.output = self.args.output
        # self.rrna = self.args.rrna
        self.SENTINEL = 1  # progressbar signal

    def load_model(self):
        """Load the right model file for classification 

        Raises:
            RuntimeError: raise error if input sequence length <40
        """

        self.len = self.args.len

        if self.len < 40:
            #             self.logger.error('{}Sequence length is too short to classify!{}'.format(
            #                 colors.FAIL,
            #                 colors.ENDC))
            #             raise RuntimeError(
            #                 "Sequence length must be set to larger than 40.")
            self.logger.info(
                'The accuracy will drop with reads shorter than 40.')

        # High recall model if ensure non-rRNA
        if self.args.ensure == 'norrna':
            model_file_ext = 'recall'
        else:
            model_file_ext = 'mcc'

        self.model_file = os.path.join(
            cd, self.config['state_file'][model_file_ext]).replace('.pth', '.onnx')

        # self.logger.info('Using high {} model file: {}{}{}{} on CPU'.format(model_file_ext.upper(),
        #                                                                     colors.BOLD,
        #                                                                     colors.OKCYAN,
        #                                                                     self.model_file,
        #                                                                     colors.ENDC))
        self.logger.info('Using high {} model'.format(model_file_ext.upper()))
        
        self.logger.info('Log file: {}'.format(
            self.args.log
            ))

        so = onnxruntime.SessionOptions()
        so.intra_op_num_threads = 1
        so.inter_op_num_threads = 1
        # so.so.enable_mem_pattern = False
        # so.enable_cpu_mem_arena = False
        # so.execution_mode = onnxruntime.ExecutionMode.ORT_SEQUENTIAL
        so.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL

        self.model = onnxruntime.InferenceSession(self.model_file, so)

    def run(self):
        """
        Load data and run the predictor
        """

        num_workers = self.args.threads

        # Manager for multiprocessing
        manager = mp.Manager()

        # List to store prediction results
        results = manager.list()
        work = manager.Queue(num_workers)

        pool = []

        # queue for progressbar signal
        q_pbar = mp.Queue()
        
        num_nonrrna = 0
        num_rrna = 0
        
        if self.is_paired:
            # Load paired end read files with multiprocessing
            with mp.Pool(2) as p:
                input_reads = p.map(SeqEncoder.load_reads, self.input)

            num_seqs = len(input_reads[0])

            self.num_batches = math.ceil(
                num_seqs / (self.batch_size))

            self.logger.info('{}{}{}{} sequences loaded!'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_seqs,
                colors.ENDC))

            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))
                rrna1_fh = open_for_write(self.rrna[0])
                rrna2_fh = open_for_write(self.rrna[1])

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))

            norrna1_fh = open_for_write(self.output[0])
            norrna2_fh = open_for_write(self.output[1])

            if self.args.ensure == 'both':

                unclf1 = self.output[0] + '.unclassified.gz'
                unclf2 = self.output[1] + '.unclassified.gz'
                unclf1_fh = open_for_write(unclf1)
                unclf2_fh = open_for_write(unclf2)
                self.logger.info('Writing unclassified sequences into file: {}{}, {}{}'.format(
                    colors.OKYELLOW,
                    unclf1,
                    unclf2,
                    colors.ENDC))

                num_unknown = 0

            # Start the listener process to minitor the progress
            proc = mp.Process(target=self.listener, args=(q_pbar,))
            proc.start()

            # Start the classification processes
            for _i in range(num_workers):
                p = mp.Process(target=self.classify_paired_reads,
                               args=(work, results, q_pbar))
                p.start()
                pool.append(p)

            # Input reads batches
            iters = itertools.chain(Predictor.generate_paired_read_batches(
                input_reads, self.batch_size), (None,) * num_workers)
            for read in iters:
                work.put(read)

            for p in pool:
                p.join()

            q_pbar.put(None)
            proc.join()

            self.logger.info('{}Writing outputs...{}'.format(
                colors.OKBLUE,
                colors.ENDC))

            for r1_dict, r2_dict in results:
                # Load the prediciton results and split the input reads accordingly
                
                num_nonrrna += len(r1_dict[0])
                num_rrna += len(r1_dict[1])

                if r1_dict[0]:
                    norrna1_fh.write('\n'.join(r1_dict[0]) + '\n')
                    norrna2_fh.write('\n'.join(r2_dict[0]) + '\n')
                if self.rrna is not None and r1_dict[1]:
                    rrna1_fh.write('\n'.join(r1_dict[1]) + '\n')
                    rrna2_fh.write('\n'.join(r2_dict[1]) + '\n')

                if self.args.ensure == 'both' and r1_dict[-1]:
                    unclf1_fh.write('\n'.join(r1_dict[-1]) + '\n')
                    unclf2_fh.write('\n'.join(r2_dict[-1]) + '\n')
                    num_unknown += len(r1_dict[-1])

                    # del r1_data, r2_data, r1_output, r2_output, r1_batch_labels, r2_batch_labels
            
            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_seqs,
                        colors.ENDC))

            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))

            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))

            if self.rrna is not None:
                rrna1_fh.close()
                rrna2_fh.close()

            if self.args.ensure == 'both':
                self.logger.info('Discarded {}{}{}{} unclassified sequences'.format(
                    colors.BOLD,
                    colors.OKCYAN,
                    num_unknown,
                    colors.ENDC))

                unclf1_fh.close()
                unclf2_fh.close()

            norrna1_fh.close()
            norrna2_fh.close()

        else:
            input_reads = SeqEncoder.load_reads(*self.input)

            num_seqs = len(input_reads)

            self.num_batches = math.ceil(
                num_seqs / (self.batch_size))

            self.logger.info('{}{}{}{} sequences loaded!'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_seqs,
                colors.ENDC))

            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))

                rrna_fh = open_for_write(self.rrna[0])
#                num_rrna = 0

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))

            norrna_fh = open_for_write(self.output[0])

            proc = mp.Process(target=self.listener, args=(q_pbar,))
            proc.start()

            for _i in range(num_workers):
                p = mp.Process(target=self.classify_reads,
                               args=(work, results, q_pbar))
                p.start()
                pool.append(p)

            iters = itertools.chain(Predictor.generate_read_batches(
                input_reads, self.batch_size), (None,) * num_workers)
            for read in iters:
                work.put(read)

            for p in pool:
                p.join()

            q_pbar.put(None)
            proc.join()

            self.logger.info('{}Writing outputs...{}'.format(
                colors.OKBLUE,
                colors.ENDC))

            for r_dict in results:

                num_nonrrna += len(r_dict[0])
                num_rrna += len(r_dict[1])
                if r_dict[0]:
                    norrna_fh.write('\n'.join(r_dict[0]) + '\n')
                if self.rrna is not None and r_dict[1]:
                    rrna_fh.write('\n'.join(r_dict[1]) + '\n')

            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_seqs,
                        colors.ENDC))
            
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))

            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))

            if self.rrna is not None:
                rrna_fh.close()
            norrna_fh.close()

    def run_with_chunks(self):
        """
        Load data with chunks and run the predictor
        """

        num_workers = self.args.threads

        read_chunk_size = self.batch_size * self.chunk_size

        num_read = 0
        num_nonrrna = 0
        num_rrna = 0

        if self.is_paired:
            self.logger.info('Classify paired-end reads with chunk size {}{}{}'.format(
                colors.BOLD,
                self.chunk_size,
                colors.ENDC))
            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))
                rrna1_fh = open_for_write(self.rrna[0])
                rrna2_fh = open_for_write(self.rrna[1])
                # num_rrna = 0

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))

            norrna1_fh = open_for_write(self.output[0])
            norrna2_fh = open_for_write(self.output[1])

            if self.args.ensure == 'both':

                unclf1 = self.output[0] + '.unclassified.gz'
                unclf2 = self.output[1] + '.unclassified.gz'
                unclf1_fh = open_for_write(unclf1)
                unclf2_fh = open_for_write(unclf2)
                self.logger.info('Writing unclassified sequences into file: {}{}, {}{}'.format(
                    colors.OKYELLOW,
                    unclf1,
                    unclf2,
                    colors.ENDC))

                num_unknown = 0

            # num_read = 0

            # Load paired end reads with chunks
            for chunk in SeqEncoder.get_pairedread_chunks(*self.input,
                                                          chunk_size=read_chunk_size):
                # Manager for multiprocessing
                manager = mp.Manager()

                # List to store prediction results
                results = manager.list()
                work = manager.Queue(num_workers)

                pool = []

                # Start the classification processes
                for _i in range(num_workers):
                    p = mp.Process(target=self.classify_paired_reads,
                                   args=(work, results))
                    p.start()
                    pool.append(p)

                # Input reads batches for each chunk
                iters = itertools.chain(Predictor.generate_paired_read_batches(
                    chunk, self.batch_size), (None,) * num_workers)
                for read in iters:
                    work.put(read)

                for p in pool:
                    p.join()

                for r1_dict, r2_dict in results:
                    # Load the prediciton results and split the input reads accordingly

                    num_nonrrna += len(r1_dict[0])
                    num_rrna += len(r1_dict[1])
                    
                    if r1_dict[0]:
                        norrna1_fh.write('\n'.join(r1_dict[0]) + '\n')
                        norrna2_fh.write('\n'.join(r2_dict[0]) + '\n')
                    if self.rrna is not None and r1_dict[1]:
                        rrna1_fh.write('\n'.join(r1_dict[1]) + '\n')
                        rrna2_fh.write('\n'.join(r2_dict[1]) + '\n')
                        # num_rrna += len(r1_dict[1])

                    if self.args.ensure == 'both' and r1_dict[-1]:
                        unclf1_fh.write('\n'.join(r1_dict[-1]) + '\n')
                        unclf2_fh.write('\n'.join(r2_dict[-1]) + '\n')
                        num_unknown += len(r1_dict[-1])

                        # del r1_data, r2_data, r1_output, r2_output, r1_batch_labels, r2_batch_labels
                num_read += len(chunk[0])

                self.logger.info('{}{}{} sequences finished!'.format(
                    colors.OKGREEN,
                    num_read,
                    colors.ENDC))

            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_read,
                        colors.ENDC))
            
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))
            
            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))

            if self.rrna is not None:
                rrna1_fh.close()
                rrna2_fh.close()

            if self.args.ensure == 'both':
                self.logger.info('Discarded {}{}{}{} unclassified sequences'.format(
                    colors.BOLD,
                    colors.OKCYAN,
                    num_unknown,
                    colors.ENDC))

                unclf1_fh.close()
                unclf2_fh.close()

            norrna1_fh.close()
            norrna2_fh.close()

        else:
            # num_read = 0
            self.logger.info('Classify reads with chunk size {}{}{}'.format(
                colors.BOLD,
                self.chunk_size,
                colors.ENDC))

            if self.rrna is not None:
                self.logger.info('Writing output rRNA sequences into file: {}{}{}'.format(
                    colors.OKBLUE,
                    ", ".join(self.rrna),
                    colors.ENDC))

                rrna_fh = open_for_write(self.rrna[0])
                # num_rrna = 0

            self.logger.info('Writing output non-rRNA sequences into file: {}{}{}'.format(
                colors.OKBLUE,
                ", ".join(self.output),
                colors.ENDC))

            norrna_fh = open_for_write(self.output[0])

            for chunk in SeqEncoder.get_seq_chunks(*self.input,
                                                   chunk_size=read_chunk_size):

                # Manager for multiprocessing
                manager = mp.Manager()

                # List to store prediction results
                results = manager.list()
                work = manager.Queue(num_workers)

                pool = []

                for _i in range(num_workers):
                    p = mp.Process(target=self.classify_reads,
                                   args=(work, results))
                    p.start()
                    pool.append(p)

                iters = itertools.chain(Predictor.generate_read_batches(
                    chunk, self.batch_size), (None,) * num_workers)
                for read in iters:
                    work.put(read)

                for p in pool:
                    p.join()

                for r_dict in results:
                    
                    num_nonrrna += len(r_dict[0])
                    num_rrna += len(r_dict[1])

                    if r_dict[0]:
                        norrna_fh.write('\n'.join(r_dict[0]) + '\n')
                    if self.rrna is not None and r_dict[1]:
                        rrna_fh.write('\n'.join(r_dict[1]) + '\n')

                num_read += len(chunk)

                self.logger.info('{}{}{} sequences finished!'.format(
                    colors.OKGREEN,
                    num_read,
                    colors.ENDC))

            self.logger.info('Processed {}{}{}{} sequences in total'.format(
                        colors.BOLD,
                        colors.OKCYAN,
                        num_read,
                        colors.ENDC))
            
            self.logger.info('Detected {}{}{}{} non-rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_nonrrna,
                colors.ENDC
            ))

            self.logger.info('Detected {}{}{}{} rRNA sequences'.format(
                colors.BOLD,
                colors.OKCYAN,
                num_rrna,
                colors.ENDC
            ))
            
            if self.rrna is not None:
                rrna_fh.close()
            norrna_fh.close()

    def detect(self):
        """Wrapper for all steps

        Raises:
            RuntimeError: raise error if the number of input read files and output files are invalid
        """
        self.input = self.args.input
        self.output = self.args.output
        self.rrna = self.args.rrna

        num_inputs = len(self.input)
        num_rrna_outputs = None if self.rrna is None else len(self.rrna)
        num_norrna_outputs = len(self.output)
        if num_inputs != num_norrna_outputs or num_inputs > 2:
            self.logger.error('{}The number of input and output sequence files is invalid!{}'.format(
                colors.FAIL,
                colors.ENDC))
            raise RuntimeError(
                "Input or output should have no more than two files and they should have the same number of files.")
        if num_rrna_outputs is not None and num_rrna_outputs != num_inputs:
            self.logger.error('{}The number of output rRNA sequence files is invalid!{}'.format(
                colors.FAIL,
                colors.ENDC))
            raise RuntimeError(
                "Ouput rRNA should have no more than two files and they should the same number with input files.")

        self.batch_size = 1024
        self.is_paired = (num_inputs == 2)

        if self.chunk_size is None:
            self.run()
        else:
            self.run_with_chunks()

    @staticmethod
    def generate_read_batches(reads, n):
        """Yield successive n-sized batches from read list."""
        for i in range(0, len(reads), n):
            yield reads[i:i + n]

    @staticmethod
    def generate_paired_read_batches(reads, n):
        """Yield successive n-sized batches from paired end read list."""
        r1, r2 = reads
        for i in range(0, len(r1), n):
            yield r1[i:i + n], r2[i:i + n]

    @staticmethod
    def separate_reads(reads, labels):
        """Split the input reads based on the predicted label

        Args:
            reads (list): batch of input reads 
            labels (list): list of predicted labels for the input reads

        Returns:
            dict: dict with key being label and value being read
        """

        reads_dict = defaultdict(list)
        for read, label in zip(reads, labels):

            reads_dict[label].append('\n'.join(read))
        return reads_dict

    def separate_paired_reads(self, r1_reads, r1_outs, r2_reads, r2_outs):
        r1_dict = defaultdict(list)
        r2_dict = defaultdict(list)

        if self.args.ensure == 'rrna':
            r1_labels = np.argmax(r1_outs, axis=1)
            r2_labels = np.argmax(r2_outs, axis=1)
            for r1, r1_label, r2, r2_label in zip(r1_reads, r1_labels, r2_reads, r2_labels):

                if r1_label == r2_label == 1:
                    final_label = 1
                else:
                    final_label = 0
                r1_dict[final_label].append('\n'.join(r1))
                r2_dict[final_label].append('\n'.join(r2))
        elif self.args.ensure == 'norrna':
            # for r1, r1_label, r2, r2_label in zip(r1, r1_labels, r2, r2_labels):
            r1_labels = np.argmax(r1_outs, axis=1)
            r2_labels = np.argmax(r2_outs, axis=1)
            for r1, r1_label, r2, r2_label in zip(r1_reads, r1_labels, r2_reads, r2_labels):
                if r1_label == r2_label == 0:
                    final_label = 0
                else:
                    final_label = 1

                r1_dict[final_label].append('\n'.join(r1))
                r2_dict[final_label].append('\n'.join(r2))
        elif self.args.ensure == 'both':
            r1_labels = np.argmax(r1_outs, axis=1)
            r2_labels = np.argmax(r2_outs, axis=1)
            for r1, r1_label, r2, r2_label in zip(r1_reads, r1_labels, r2_reads, r2_labels):
                # for r1, r1_label, r2, r2_label in zip(r1, r1_labels, r2, r2_labels):
                if r1_label == r2_label == 0:
                    final_label = 0
                elif r1_label == r2_label == 1:
                    final_label = 1
                else:
                    final_label = -1

                r1_dict[final_label].append('\n'.join(r1))
                r2_dict[final_label].append('\n'.join(r2))
        else:

            final_labels = np.argmax(r1_outs + r2_outs, axis=1)
            for r1, r2, final_label in zip(r1_reads, r2_reads, final_labels):

                r1_dict[final_label].append('\n'.join(r1))
                r2_dict[final_label].append('\n'.join(r2))

        return r1_dict, r2_dict

    def classify_reads(self, batch, out_list, q_pbar=None):
        """Classify reads batch

        Args:
            batch (list): batch of reads
            out_list (list(dict)): list of separated {label: read} dicts
            q_pbar (mp queue): queue for progressbar signal
        """
        while True:
            reads = batch.get()
            if reads == None:
                return

            input_encoded_reads = np.array([SeqEncoder.encode_variable_len_read(
                read[1], max_len=self.len) for read in reads], dtype=np.float32)

            inputs = {self.model.get_inputs()[0].name: input_encoded_reads}
            outputs = self.model.run(None, inputs)

            out_list.append(Predictor.separate_reads(
                reads, np.argmax(outputs[0], axis=1)))
            if q_pbar is not None:
                q_pbar.put(self.SENTINEL)

    def classify_paired_reads(self, batch, out_list, q_pbar=None):
        while True:
            reads = batch.get()
            if reads == None:
                return

            r1, r2 = reads

            input_encoded_r1 = np.array([SeqEncoder.encode_variable_len_read(
                read[1], max_len=self.len) for read in r1], dtype=np.float32)

            input_encoded_r2 = np.array([SeqEncoder.encode_variable_len_read(
                read[1], max_len=self.len) for read in r2], dtype=np.float32)

            output_r1 = self.model.run(
                None, {self.model.get_inputs()[0].name: input_encoded_r1})[0]

            output_r2 = self.model.run(
                None, {self.model.get_inputs()[0].name: input_encoded_r2})[0]

            out_list.append(self.separate_paired_reads(
                r1, output_r1, r2, output_r2))

            if q_pbar is not None:
                q_pbar.put(self.SENTINEL)

    def listener(self, q_pbar):
        pbar = tqdm(total=self.num_batches)
        for _ in iter(q_pbar.get, None):
            pbar.update()


def open_for_write(read_file):
    """Open a plain text or gzipped text file for writing according to the file name extension

    Args:
        read_file (str): the file name to write

    Returns:
        file handle: the opened file handle for writing
    """
    if read_file.endswith('gz'):
        return gzip.open(read_file, mode='wt', compresslevel=5)
    else:
        return open(read_file, 'w')


class colors:
    """
    Define the color of logger text
    """

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    OKYELLOW = '\033[33m'
    OKMAG = '\033[35m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    UPDATE = '\033[F'


def main():
    args = argparse.ArgumentParser(
        description='rRNA sequence detector', formatter_class=RawTextHelpFormatter)
    args.add_argument('-c', '--config', default=None, type=str,
                      help='Path of config file')

    args.add_argument('-l', '--len', type=int, required=True,
                      help='Sequencing read length. Note: the accuracy reduces for reads shorter than 40.')
    args.add_argument('-i', '--input', default=None, type=str, nargs='*', required=True,
                      help='Path of input sequence files (fasta and fastq), the second file will be considered as second end if two files given.')
    args.add_argument('-o', '--output', default=None, type=str, nargs='*', required=True,
                      help='Path of the output sequence files after rRNAs removal (same number of files as input). \n(Note: 2 times slower to write gz files)')
    args.add_argument('-r', '--rrna', default=None, type=str, nargs='*',
                      help='Path of the output sequence file of detected rRNAs (same number of files as input)')
    args.add_argument('-e', '--ensure', default="none", type=str, choices=['rrna', 'norrna', 'both', 'none'],
                      help='''Ensure which classificaion has high confidence for paired end reads.
norrna: output only high confident non-rRNAs, the rest are clasified as rRNAs;
rrna: vice versa, only high confident rRNAs are classified as rRNA and the rest output as non-rRNAs;
both: both non-rRNA and rRNA prediction with high confidence;
none: give label based on the mean probability of read pair.
      (Only applicable for paired end reads, discard the read pair when their predicitons are discordant)''')

    args.add_argument('-t', '--threads', default=20, type=int,
                      help='Number of threads to use. (default: 20)')

    args.add_argument('--chunk_size', default=None, type=int,
                      help='chunk_size * 1024 reads to load each time. \n{}.'.format(
                          'When chunk_size=1000 and threads=20, consumming ~20G memory, better to be multiples of the number of threads.'))
    args.add_argument('--log', default=None, type=str,
                      help='Log file name')
    args.add_argument('-v', '--version', action='version',
                      version='%(prog)s {version}'.format(version=__version__))

    if not isinstance(args, tuple):
        args = args.parse_args()
    if args.config is None:
        config_file = os.path.join(cd, 'config.json')
    else:
        config_file = args.config
    config = ConfigParser.from_json(config_file)

    os.environ['OMP_NUM_THREADS'] = '1'
    # os.environ['MKL_NUM_THREADS'] = '1'

    if platform.system() == 'Darwin':
        mp.set_start_method('fork')
    seq_pred = Predictor(config, args)
    seq_pred.load_model()

    seq_pred.detect()


if __name__ == '__main__':
    main()



'''
File: fastx_parser.py
Created Date: November 30th 2019
Author: ZL Deng <dawnmsg(at)gmail.com>
---------------------------------------
Last Modified: 30th November 2019 11:04:48 pm

Description:

This is a simple fastq and fasta parser which is about 5 times faster than
SeqIO in Biopython.
'''


def seq_parser(seq_fh, seq_type):
    if seq_type == 'fastq':
        record_line = 0  # before parse seq record
        for line in seq_fh:
            line = line.rstrip()
            # if the current line is header line
            if line[0] == '@' and record_line == 0:
                record_line = 1
                header = line
            # if it is the next line of the header which is the seq reads
            elif record_line == 1:
                seq = line
                record_line = 2
            # if it is the next line of seq, which is + line
            elif record_line == 2:
                endseq = line
                record_line = 3
            # quality line
            else:
                qual = line
                yield header, seq, endseq, qual
                header = seq = endseq = qual = ''
                record_line = 0

    else:
        seq = ''
        header = ''
        for line in seq_fh:
            line = line.strip()
            if line != '':
                if line[0] == '>':
                    if header != '':
                        yield header, seq
                        header = line
                        seq = ''
                    else:
                        header = line
                else:
                    seq += line.upper()
        if seq != '' and seq is not None:
            yield header, seq



# from data_loader.fastx_parser import seq_parser
from torch.utils.data import Dataset, IterableDataset, get_worker_info

# Single end sequence data, and long sequence data


class SeqData(Dataset):
    def __init__(self, data):
        self.data = data

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)


# Data is paired end read list with two items
class PairedReadData(Dataset):
    def __init__(self, data):
        self.data = data

    def __getitem__(self, index):
        r1 = self.data[0][index]
        r2 = self.data[1][index]
        return r1, r2

    def __len__(self):
        return len(self.data[0])


class IterablePairedReadData(IterableDataset):
    '''
    @params:
        r1, r2: read end pairs
    '''

    def __init__(self, r1, r2, batch_size):
        # super(IterablePairedReadData).__init__()
        self.r1 = r1
        self.r2 = r2
        self.batch_size = batch_size

    def get_stream(self):
        for (r1_read, r2_read) in zip(self.r1, self.r2):
            worker = get_worker_info()
            worker_id = id(self) if worker is not None else -1
            yield (r1_read, r2_read), worker_id

    def __iter__(self):
        return zip(*[self.get_stream for _ in range(self.batch_size)])

    @classmethod
    def split_data(cls, r1, r2, batch_size, max_workers):
        # Determine the proper number of workers
        for n in range(max_workers, 0, -1):
            if batch_size % n == 0:
                num_workers = n
                break
        split_size = batch_size // num_workers
        return [cls(r1, r2, batch_size=split_size)]


class IterableSeqData(IterableDataset):
    '''
    @params:
        data: read generator
    '''

    def __init__(self, data, batch_size):
        # super(IterablePairedReadData).__init__()
        self.data = data
        self.batch_size = batch_size

    def get_stream(self):
        for seq in self.data:
            worker = get_worker_info()
            worker_id = id(self) if worker is not None else -1
            yield seq, worker_id

    # def get_batch(self):
    #     return zip(*[self.get_stream for _ in range(self.batch_size)])

    def __iter__(self):
        return zip(*[self.get_stream for _ in range(self.batch_size)])

    @classmethod
    def split_data(cls, data, batch_size, max_workers):
        # Determine the proper number of workers
        for n in range(max_workers, 0, -1):
            if batch_size % n == 0:
                num_workers = n
                break
        split_size = batch_size // num_workers
        return [cls(data, batch_size=split_size)]



from multiprocessing.reduction import ForkingPickler, AbstractReducer


class ForkingPickler4(ForkingPickler):
    def __init__(self, *args):
        if len(args) > 1:
            args[1] = 2
        else:
            args.append(2)
        super().__init__(*args)

    @classmethod
    def dumps(cls, obj, protocol=4):
        return ForkingPickler.dumps(obj, protocol)


def dump(obj, file, protocol=4):
    ForkingPickler4(file, protocol).dump(obj)


class Pickle4Reducer(AbstractReducer):
    ForkingPickler = ForkingPickler4
    register = ForkingPickler4.register
    dump = dump



import torch
import math


def accuracy(output, target):
    with torch.no_grad():
        pred = torch.argmax(output, dim=1)
        assert pred.shape[0] == len(target)
        correct = 0
        correct += torch.sum(pred == target).item()
    return correct / len(target)


def all(output, target):
    tn = tp = fn = fp = 0
    with torch.no_grad():
        pred = torch.argmax(output, dim=1)
        #print("Metric:", output[:5])
        #print("argmax:", y_pred[:5])
        # pred = torch.max(output, 1)[1]  # .numpy().squeeze()
        #print("max:", pred[:5])
        assert pred.shape[0] == len(target)

        for (i, j) in zip(pred, target):
            i = int(i)
            j = int(j)
            #print(i, j)
            if i == 0 and j == 0:
                tn += 1
            elif i == 1 and j == 1:
                tp += 1
            elif i == 0 and j == 1:
                fn += 1
            else:
                fp += 1
        #print(tn, tp, fn, fp)
        try:
            recall = tp / (tp + fn)
        except ZeroDivisionError:
            recall = 0
        try:
            precision = tp / (tp + fp)
        except ZeroDivisionError:
            precision = 0
        accuracy = (tp + tn) / (tp + fp + tn + fn)
        try:
            F1 = 2 * (recall * precision) / (recall + precision)
        except ZeroDivisionError:
            F1 = 0
        try:
            mcc = (tp * tn - fp * fn) / math.sqrt((tp + fp)
                                                  * (tp + fn) * (tn + fp) * (tn + fn))
        except ZeroDivisionError:
            mcc = 0

    return recall, precision, accuracy, F1, mcc


def top_k_acc(output, target, k=2):
    with torch.no_grad():
        pred = torch.topk(output, k, dim=1)[1]
        assert pred.shape[0] == len(target)
        correct = 0
        for i in range(k):
            correct += torch.sum(pred[:, i] == target).item()
    return correct / len(target)


def recall(output, target):
    return all(output, target)[0]


def precision(output, target):
    return all(output, target)[1]


def F1(output, target):
    return all(output, target)[2]


def mcc(output, target):
    return all(output, target)[3]



import logging
import logging.config
from pathlib import Path
from utils import read_json


def setup_logging(save_dir, log_config='logger/logger_config.json', default_level=logging.INFO):
    """
    Setup logging configuration
    """
    log_config = Path(log_config)
    if log_config.is_file():
        config = read_json(log_config)
        # modify logging paths based on run config
        for _, handler in config['handlers'].items():
            if 'filename' in handler:
                handler['filename'] = str(save_dir / handler['filename'])

        logging.config.dictConfig(config)
    else:
        print("Warning: logging configuration file is not found in {}.".format(log_config))
        logging.basicConfig(level=default_level)



import torch
from torch import jit
import torch.nn as nn
from typing import Tuple
from torch import Tensor
from ribodetector.base import BaseModel
from torch.nn.utils.rnn import PackedSequence


class SeqModel(BaseModel):
    def __init__(self, input_size, hidden_size, num_layers, num_classes,
                 batch_first=True,
                 bidirectional=True,
                 pack_seq=True):
        super(SeqModel, self).__init__()
        self.rnn = nn.LSTM(
            input_size=input_size,      # width of input
            hidden_size=hidden_size,     # number of rnn hidden unit
            num_layers=num_layers,       # number of RNN layers
            batch_first=batch_first,
            bidirectional=bidirectional,
        )

        self.out = nn.Linear(hidden_size * 2, num_classes)    # output layer

        if pack_seq:
            self.forward = self.forward1
        else:
            self.forward = self.forward2

    # packedsequence input, extract last unpad output
    def forward1(self, x):
        r_out, _ = self.rnn(x, None)
        last_out = last_items(pack=r_out, unsort=True)

        out = self.out(last_out)
        return out

    # paddedsequence input, extract last unpad output
    def forward2(self, x):
        r_out, _ = self.rnn(x, None)
        # last_out = r_out[torch.arange(
        #     x.size(0)), (x.sum(2) != 0).cumsum(1).argmax(1), :]
        # last_out = r_out[torch.arange(x.size(0)), x.size(
        #     1) - 1 - (torch.flip(x.sum(2), [1])).argmax(1), :]

        last_out = last_pad_out_items(x, r_out)

        out = self.out(last_out)
        return out

    # paddedsequence input, extract last output no matter if they are from padded step
    def forward3(self, x):
        r_out, _ = self.rnn(x, None)

        out = self.out(r_out[:, -1, :])
        return out

    # paddedsequence or packedsequence input, extract h_n output
    def forward4(self, x):
        _r_out, (h_n, _c_n) = self.rnn(x, None)
        last_out = torch.cat((h_n[0], h_n[1]), 1)
        out = self.out(last_out)
        return out


@jit.script
def last_pad_out_items(ts: Tensor, r_out: Tensor) -> Tensor:
    last_out = r_out[torch.arange(ts.size(0)), ts.size(
        1) - 1 - ts.sum(2).flip(1).argmax(1), :]

    return last_out


@jit.script
def sorted_lengths(pack: PackedSequence) -> Tuple[Tensor, Tensor]:
    indices = torch.arange(
        pack.batch_sizes[0],
        dtype=pack.batch_sizes.dtype,
        device=pack.batch_sizes.device,
    )
    lengths = ((indices + 1)[:, None] <=
               pack.batch_sizes[None, :]).long().sum(dim=1)
    return lengths, indices


@jit.script
def sorted_first_indices(pack: PackedSequence) -> Tensor:
    return torch.arange(
        pack.batch_sizes[0],
        dtype=pack.batch_sizes.dtype,
        device=pack.batch_sizes.device,
    )


@jit.script
def sorted_last_indices(pack: PackedSequence) -> Tensor:
    lengths, indices = sorted_lengths(pack)
    cum_batch_sizes = torch.cat([
        pack.batch_sizes.new_zeros((2,)),
        torch.cumsum(pack.batch_sizes, dim=0),
    ], dim=0)
    return cum_batch_sizes[lengths] + indices


@jit.script
def first_items(pack: PackedSequence, unsort: bool) -> Tensor:
    if unsort and pack.unsorted_indices is not None:
        return pack.data[pack.unsorted_indices]
    else:
        return pack.data[:pack.batch_sizes[0]]


@jit.script
def last_items(pack: PackedSequence, unsort: bool) -> Tensor:
    indices = sorted_last_indices(pack=pack)
    if unsort and pack.unsorted_indices is not None:
        indices = indices[pack.unsorted_indices]
    return pack.data[indices]



import logging
from pathlib import Path
from functools import reduce, partial
from operator import getitem
from datetime import datetime
# from logger import setup_logging
from ribodetector.utils import read_json, write_json


class ConfigParser:
    def __init__(self, config):
        """
        class to parse configuration json file. Handles hyperparameters for training, initializations of modules, checkpoint saving
        and logging module.
        :param config: Dict containing configurations, hyperparameters for training. contents of `config.json` file for example.
        :param resume: String, path to the checkpoint being loaded.
        :param modification: Dict keychain:value, specifying position values to be replaced from config dict.
        :param run_id: Unique Identifier for training processes. Used to save checkpoints and training log. Timestamp is being used as default
        """
        # load config file and apply modification
        self.config = config

        self.log_levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG
        }

    @classmethod
    def from_json(cls, config_json):
        """
        Initialize this class from some cli arguments. Used in train, test.
        """
        # if not isinstance(args, tuple):
        #     args = args.parse_args()

        cfg_fname = Path(config_json)
        config = read_json(cfg_fname)
        # print(config)
        # print(cls(config))
        return cls(config)

    def init_obj(self, name, module, *args, **kwargs):
        """
        Finds a function handle with the name given as 'type' in config, and returns the
        instance initialized with corresponding arguments given.

        `object = config.init_obj('name', module, a, b=1)`
        is equivalent to
        `object = module.name(a, b=1)`
        """
        module_name = self[name]['type']
        module_args = dict(self[name]['args'])
        assert all([k not in module_args for k in kwargs]
                   ), 'Overwriting kwargs given in config file is not allowed'
        module_args.update(kwargs)
        return getattr(module, module_name)(*args, **module_args)

    def init_ftn(self, name, module, *args, **kwargs):
        """
        Finds a function handle with the name given as 'type' in config, and returns the
        function with given arguments fixed with functools.partial.

        `function = config.init_ftn('name', module, a, b=1)`
        is equivalent to
        `function = lambda *args, **kwargs: module.name(a, *args, b=1, **kwargs)`.
        """
        module_name = self[name]['type']
        module_args = dict(self[name]['args'])
        assert all([k not in module_args for k in kwargs]
                   ), 'Overwriting kwargs given in config file is not allowed'
        module_args.update(kwargs)
        return partial(getattr(module, module_name), *args, **module_args)

    def __getitem__(self, name):
        """Access items like ordinary dict."""
        return self.config[name]

    def get_logger(self, name, verbosity=2, logfile=None):
        handlers = [logging.StreamHandler()]
        if logfile is not None:
            handlers.append(logging.FileHandler(logfile, mode='w'))
        msg_verbosity = 'verbosity option {} is invalid. Valid options are {}.'.format(
            verbosity, self.log_levels.keys())
        assert verbosity in self.log_levels, msg_verbosity
        logging.basicConfig(
            level=self.log_levels[verbosity],
            format='%(asctime)s : %(levelname)s  %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            handlers=handlers)
        logger = logging.getLogger(name)
        return logger

    # # setting read-only attributes
    # @property
    # def config(self):
    #     return self._config

    # @property
    # def save_dir(self):
    #     return self._save_dir

    # @property
    # def log_dir(self):
    #     return self._log_dir


# helper functions to update config dict with custom cli options
# def _update_config(config, modification):
#     if modification is None:
#         return config

#     for k, v in modification.items():
#         if v is not None:
#             _set_by_path(config, k, v)
#     return config


def _get_opt_name(flags):
    for flg in flags:
        if flg.startswith('--'):
            return flg.replace('--', '')
    return flags[0].replace('--', '')


def _set_by_path(tree, keys, value):
    """Set a value in a nested object in tree by sequence of keys."""
    keys = keys.split(';')
    _get_by_path(tree, keys[:-1])[keys[-1]] = value


def _get_by_path(tree, keys):
    """Access a nested object in tree by sequence of keys."""
    return reduce(getitem, keys, tree)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from setuptools import find_packages, setup

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

required = [
    "pandas",
    "tqdm",
    "numpy",
    "biopython",
    "onnxruntime >= 1.10.0, <= 1.15.1",
    "torch >= 1.7.1, <= 1.12.1",
]

setup(
    name="ribodetector",
    version="0.3.1",
    python_requires=">=3.8, <=3.12",
    author="Z-L Deng",
    author_email="dawnmsg@gmail.com",
    description="Accurate and rapid RiboRNA sequences Detector based on deep learning.",
    license="GPL-3 License",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hzi-bifo/RiboDetector",
    packages=find_packages(include=["ribodetector", "ribodetector.*"]),
    package_data={'': ['*.json', '*.yaml', '*.pth', '*.onnx']},

    include_package_data=True,

    classifiers=[
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'ribodetector = ribodetector.detect:main',
            'ribodetector_cpu = ribodetector.detect_cpu:main',
        ]
    },
    zip_safe=True,
    install_requires=required
)



# import itertools
import gzip
from Bio.Seq import Seq
from pathlib import Path
from functools import partial
from itertools import islice
from mimetypes import guess_type
from ribodetector.data_loader.fastx_parser import seq_parser


BASE_DICT = {"A": (1, 0, 0, 0),
             "C": (0, 1, 0, 0),
             "G": (0, 0, 1, 0),
             "T": (0, 0, 0, 1),
             "U": (0, 0, 0, 1)
             }

ZERO_LIST = (0, 0, 0, 0)


def get_seq_format(seq_file):
    fa_exts = [".fasta", ".fa", ".fna", ".fas"]
    fq_exts = [".fq", ".fastq"]
    encoding = guess_type(seq_file)[1]  # uses file extension
    if encoding is None:
        encoding = ""
    elif encoding == "gzip":
        encoding = "gz"
    else:
        raise ValueError('Unknown file encoding: "{}"'.format(encoding))
    seq_filename = Path(
        seq_file).stem if encoding == 'gz' else Path(seq_file).name
    seq_file_ext = Path(seq_filename).suffix
    if seq_file_ext not in (fa_exts + fq_exts):
        raise ValueError("""Unknown extension {}. Only fastq and fasta sequence formats are supported.
And the file must end with one of ".fasta", ".fa", ".fna", ".fas", ".fq", ".fastq"
and followed by ".gz" or ".gzip" if they are gzipped.""".format(seq_file_ext))
    seq_format = "fa" + encoding if seq_file_ext in fa_exts else "fq" + encoding
    return seq_format


def load_seqs(seq_file, label, step_size):
    seq_step_list = []
    seq_format = get_seq_format(seq_file)
    _open = partial(gzip.open, mode='rt') if seq_format.endswith(
        "gz") else open
    seq_type = "fasta" if seq_format.startswith("fa") else "fastq"
    with _open(seq_file) as fh:
        for record in seq_parser(fh, seq_type):
            seq = record[1]
            seq_step_list.extend([(label, seq, step_size),
                                  (label, str(Seq(seq).reverse_complement()), step_size)])
    return seq_step_list


def load_reads(seq_file, label=None, max_len=100):
    read_list = []
    seq_format = get_seq_format(seq_file)
    _open = partial(gzip.open, mode='rt') if seq_format.endswith(
        "gz") else open
    seq_type = "fasta" if seq_format.startswith("fa") else "fastq"
    if label is None:
        with _open(seq_file) as fh:
            for record in seq_parser(fh, seq_type):
                read_list.append(record)
    else:
        with _open(seq_file) as fh:
            for record in seq_parser(fh, seq_type):
                seq = record[1]
                read, rc_read = get_read_rc_with_maxlen(seq, max_len=100)
                read_list.extend([(label, read), (label, rc_read)])
    return read_list


def get_seq_chunks(seq_file, chunk_size=1048576):
    seq_format = get_seq_format(seq_file)
    _open = partial(gzip.open, mode='rt') if seq_format.endswith(
        "gz") else open
    seq_type = "fasta" if seq_format.startswith("fa") else "fastq"
    with _open(seq_file) as fh:
        seq_iterator = seq_parser(fh, seq_type)
        while True:
            seqs_chunk = list(islice(seq_iterator, chunk_size))
            if seqs_chunk:
                yield seqs_chunk
            else:
                break


def get_pairedread_chunks(r1_seq_file, r2_seq_file, chunk_size=1048576):
    for r1_chunk, r2_chunk in zip(get_seq_chunks(r1_seq_file, chunk_size), get_seq_chunks(r2_seq_file, chunk_size)):
        yield r1_chunk, r2_chunk


def get_read_rc_with_maxlen(seq, max_len=100):
    read = rc_read = 'N' * max_len
    seq_len = len(seq)
    if seq_len >= max_len:
        start = (seq_len - max_len) // 2
        end = max_len + start
        read = seq[start:end]
        rc_read = str(Seq(read).reverse_complement())
    else:
        missing_len = max_len - seq_len
        read = seq + 'N' * missing_len
        rc_read = str(Seq(seq).reverse_complement()) + 'N' * missing_len

    return read, rc_read


def get_read_with_maxlen(seq, max_len=100):
    read = 'N' * max_len
    seq_len = len(seq)
    if seq_len > max_len:
        start = (seq_len - max_len) // 2
        end = max_len + start
        read = seq[start:end]
    elif seq_len == max_len:
        read = seq
    else:
        missing_len = max_len - seq_len
        read = seq + 'N' * missing_len
    return read


def encode_read(read):
    return [BASE_DICT.get(base, ZERO_LIST) for base in read]


def encode_variable_len_read(read, max_len=100):

    read_len = len(read)
    if read_len >= max_len:
        # start = (read_len - max_len) // 2
        # end = max_len + start
        # encoded_read = [BASE_DICT.get(base, ZERO_LIST) for base in read[start:end]]
        encoded_read = [BASE_DICT.get(base, ZERO_LIST)
                        for base in read[:max_len]]
    # elif read_len == max_len:
    #     encoded_read = [BASE_DICT.get(base, ZERO_LIST) for base in read]
    else:
        encoded_read = [ZERO_LIST] * max_len
        encoded_read[:read_len] = [BASE_DICT.get(
            base, ZERO_LIST) for base in read]
    return encoded_read


def encode_seq_reads(seq, step_size, max_len=100):
    seq_len = len(seq)
    seq_feature = [BASE_DICT.get(base, ZERO_LIST) for base in seq]
    encoded_read_list = []
    for i in range(0, seq_len, step_size):
        encoded_read = [ZERO_LIST] * max_len

        if seq_len >= i + max_len:
            encoded_read_list.append(seq_feature[i:i + max_len])
        else:
            if seq_len > i + max_len / 2:
                encoded_read[:seq_len - i] = seq_feature[i:]
                encoded_read_list.append(encoded_read)
            break
    return encoded_read_list
